# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import copy
from itertools import zip_longest
from vermouth.parser_utils import SectionLineParser
from vermouth.molecule import Interaction
from vermouth.gmx.itp_read import read_itp
from .meta_molecule import MetaMolecule, _make_edges
from tqdm import tqdm

class TOPDirector(SectionLineParser):

    COMMENT_CHAR = ';'

    atom_idxs = {'bonds': [0, 1],
                 'bondtypes':[0, 1],
                 'position_restraints': [0],
                 'angles': [0, 1, 2],
                 'angletypes':[0, 1, 2],
                 'constraints': [0, 1],
                 'constrainttypes': [0, 1],
                 'dihedrals': [0, 1, 2, 3],
                 'dihedraltypes': [0, 1, 2, 3],
                 'pairs': [0, 1],
                 'pairtypes': [0, 1],
                 'exclusions': [slice(None, None)],
                 'virtual_sitesn': [0, slice(2, None)],
                 'virtual_sites2': [0, 1, 2, 3],
                 'virtual_sites3': [0, 1, 2, 3],
                 'pairs_nb': [0, 1],
                 'SETTLE': [0],
                 'virtual_sites4': [slice(0, 5)],
                 'distance_restraints':  [0, 1],
                 'dihedral_restraints':  [slice(0, 4)],
                 'orientation_restraints': [0, 1],
                 'angle_restraints': [slice(0, 4)],
                 'angle_restraints_z': [0, 1]}

    def __init__(self, topology, cwdir=None):
        super().__init__()
        self.force_field = topology.force_field
        self.topology = topology
        self.current_meta = None
        self.current_itp = None
        self.itp_lines = []
        self.molecules = []
        self.cwdir = cwdir
        self.header_actions = {
            ('moleculetype',): self._new_itp
        }
        self.pragma_actions = {
            '#define': self.parse_define,
            '#include': self.parse_include
        }

    def dispatch(self, line):
        """
        Looks at `line` to see what kind of line it is, and returns either
        :meth:`parse_header` if `line` is a section header or
        :meth:`vermouth.parser_utils.SectionLineParser.parse_section` otherwise.
        Calls :meth:`vermouth.parser_utils.SectionLineParser.is_section_header` to see
        whether `line` is a section header or not.

        Parameters
        ----------
        line: str

        Returns
        -------
        collections.abc.Callable
            The method that should be used to parse `line`.
        """
        if self.is_pragma(line):
            return self.parse_top_pragma
        elif self.is_star_comment(line):
            return self._skip
        else:
            return super().dispatch(line)

    @staticmethod
    def is_pragma(line):
        """
        Parameters
        ----------
        line: str
            A line of text.

        Returns
        -------
        bool
            ``True`` if `line` is a def statement.
        """
        return line.startswith('#')

    @staticmethod
    def is_star_comment(line):
        """
        Star comments are special comments
        usually found at the beginning of GMX
        library topology files. They are different
        from the regular comment as they are specific
        to top library files.

        Parameters
        ----------
        line: str
            A line of text.

        Returns
        -------
        bool
            ``True`` if `line` is a star comment.
        """
        return line.startswith('*')

    def parse_top_pragma(self, line, lineno=0):
        """
        Parses the beginning and end of define sections
        with line number `lineno`. Sets attr current_meta
        when applicable. Does check if ifdefs overlap.

        Parameters
        ----------
        line: str
        lineno: str

        Returns
        -------
        object
               The result of calling :meth:`finalize_section`, which is called
               if a section ends.

        Raises
        ------
        IOError
            If the def sections are missformatted
        """
        if line == '#endif':
            if self.current_itp:
                self.current_itp.append(line)
            elif self.current_meta is None:
                raise IOError("Your #ifdef section is orderd incorrectly."
                              "At line {} I read {} but I haven not read"
                              "an #ifdef before.".format(lineno, line))
            else:
               self.current_meta = None

        elif line.startswith("#else"):
            if self.current_itp:
                self.current_itp.append(line)
            elif self.current_meta is None:
               raise IOError("Your #ifdef section is orderd incorrectly."
                             "At line {} I read {} but I haven not read"
                             "a ifdef before.".format(lineno, line))
            else:
               inverse = {"ifdef": "ifndef", "ifndef": "ifdef"}
               tag = self.current_meta["tag"]
               condition = inverse[self.current_meta["condition"]]
               self.current_meta = {'tag': tag, 'condition': condition}

        elif line.startswith("#ifdef") or line.startswith("#ifndef"):
            if self.current_itp:
                self.current_itp.append(line)
            elif self.current_meta is None:
                condition, tag = line.split()
                self.current_meta = {'tag': tag,
                                     'condition': condition.replace("#", "")}
            elif self.current_meta is not None:
                raise IOError("Your {} section is orderd incorrectly."
                              "At line {} I read {} but there is still"
                              "an open #ifdef/#ifndef section from"
                              "before.".format(self.current_meta['tag'], lineno, line.split()[0]))

        elif line.split()[0] in self.pragma_actions:
            action = self.pragma_actions[line.split()[0]]
            action(line)
        else:
            raise IOError("Don't know how to parse pargma {} at"
                          "line {}.".format(line, lineno))

    def parse_header(self, line, lineno=0):
        """
        Parses a section header with line number `lineno`. Sets
        :attr:`vermouth.parser_utils.SectionLineParser.section`
        when applicable. Does not check whether `line` is a valid section
        header.

        Parameters
        ----------
        line: str
        lineno: str

        Returns
        -------
        object
            The result of calling :meth:`finalize_section`, which is called
            if a section ends.

        Raises
        ------
        KeyError
            If the section header is unknown.
        """
        result = super().parse_header(line, lineno)
        action = self.header_actions.get(tuple(self.section))
        if action:
            action()

        if self.current_itp is not None:
            self.current_itp.append(line)
        return result

    def finalize(self, lineno=0):
        """
        Called at the end of the file and checks that all pragmas are closed
        before calling the parent method.
        """
        if self.current_itp:
            self.itp_lines.append(self.current_itp)

        if self.current_meta is not None:
            raise IOError("Your {} section is orderd incorrectly."
                          "There is no #endif for this pragma.".format(self.current_meta))

        for lines in self.itp_lines:
            read_itp(lines, self.force_field)

        total_count = 0
        _make_edges(self.force_field)

        for mol_name, n_mol in self.molecules:
            block = self.force_field.blocks[mol_name]
            graph = MetaMolecule._block_graph_to_res_graph(block)
            for idx in tqdm(range(0, int(n_mol))):
                graph_copy = graph.copy(as_view=False)
                new_mol = MetaMolecule(graph_copy,
                                       force_field=self.force_field,
                                       mol_name=mol_name)
                new_mol.molecule = self.force_field.blocks[mol_name].to_molecule()
                new_mol.mol_name = mol_name

                self.topology.add_molecule(new_mol)
                self.topology.mol_idx_by_name[mol_name].append(total_count)
                total_count += 1

        super().finalize()

    def _new_itp(self):
        if self.current_itp:
           self.itp_lines.append(self.current_itp)
        self.current_itp = []

    @SectionLineParser.section_parser('system')
    def _system(self, line, lineno=0):
        """
        Parses the lines in the '[system]'
        directive and stores it.
        """
        system_lines = self.topology.description
        system_lines.append(line)
        self.description = system_lines

    @SectionLineParser.section_parser('molecules')
    def _molecules(self, line, lineno=0):
        """
        Parses the lines in the '[molecules]'
        directive and stores it.
        """
        # we need to keep the order here so cannot make it a dict
        # also mol names do not need to be unique
        name, n_mol = line.split()
        self.molecules.append((name, n_mol))

    @SectionLineParser.section_parser('defaults')
    def _defaults(self, line, lineno=0):
        """
        Parse and store the defaults section.
        """
        defaults = ["nbfunc", "comb-rule", "gen-pairs", "fudgeLJ", "fudgeQQ"]
        numbered_terms = ["nbfunc", "comb-rule", "fudgeLJ", "fudgeQQ"]
        tokens = line.split()

        # Parse all defaults to a dict up to the last default metioned
        # Note that gen_pairs, fudgeLJ and fudgeQQ not need to be set
        self.topology.defaults = dict(zip(defaults[0:len(tokens)], tokens))

        # we cannot interpret Buckingham Potentials at the moment so we crash
        if self.topology.defaults["nbfunc"] == "2":
           raise IOError("Buckingham Potential requested but this potential form"
                         "currently is not implemented.")

        # converts the defaults that are numbers to numbers
        # we need to parse them first because they are not guaranteed to be provided
        for token_name in numbered_terms:
            if token_name in self.topology.defaults:
                 self.topology.defaults[token_name] = float(self.topology.defaults[token_name])

        # sets gen-pairs to no when it is not provied
        if "gen-pairs" not in self.topology.defaults:
            self.topology.defaults["gen-pairs"] = "no"

    @SectionLineParser.section_parser('atomtypes')
    def _atomtypes(self, line, lineno=0):
        """
        Parse and store atomtypes section
        """
        tokens = line.split()
        atom_name = tokens.pop(0)
        tokens.reverse()
        atom_type_line = dict(zip_longest(["nb2", "nb1", "ptype",
                                           "charge", "mass",
                                           "atom_num", "bond_type"], tokens, fillvalue=None))
        floats = ["nb1", "nb2", "charge", "mass", "atom_num"]
        for term, value in atom_type_line.items():
             if term in floats and value:
                 atom_type_line[term] = float(value)

        if None in atom_type_line:
            msg = ("Can't parse line {}. Found more parameters than expected.")
            raise OSError(msg.format(line))

        self.topology.atom_types[atom_name] = atom_type_line
        
    @SectionLineParser.section_parser('nonbond_params')
    def _nonbond_params(self, line, lineno=0):
        """
        Parse and store nonbond params
        """
        atom_1, atom_2, func, nb1, nb2 = line.split()
        self.topology.nonbond_params[frozenset([atom_1, atom_2])] = {"f": int(func),
                                                                     "nb1": float(nb1),
                                                                     "nb2": float(nb2)}
    @SectionLineParser.section_parser('pairtypes')
    @SectionLineParser.section_parser('angletypes')
    @SectionLineParser.section_parser('dihedraltypes')
    @SectionLineParser.section_parser('bondtypes')
    @SectionLineParser.section_parser('constrainttypes')
    def _type_params(self, line, lineno=0):
        """
        Parse and store bonded types
        """
        section_name = self.section[-1]
        inter_type = section_name[:-5] + "s"
        atoms, params = self._split_atoms_and_parameters(line.split(),
                                                         self.atom_idxs[section_name])

        self.topology.types[inter_type][tuple(atoms)] = (params, self.current_meta)


    @SectionLineParser.section_parser('implicit_genborn_params')
    @SectionLineParser.section_parser('cmaptypes')
    def _skip(self, line, lineno=0):
        pass

    @SectionLineParser.section_parser('moleculetype')
    @SectionLineParser.section_parser('moleculetype', 'atoms')
    @SectionLineParser.section_parser('moleculetype', 'bonds')
    @SectionLineParser.section_parser('moleculetype', 'angles')
    @SectionLineParser.section_parser('moleculetype', 'dihedrals')
    @SectionLineParser.section_parser('moleculetype', 'impropers')
    @SectionLineParser.section_parser('moleculetype', 'constraints')
    @SectionLineParser.section_parser('moleculetype', 'pairs')
    @SectionLineParser.section_parser('moleculetype', 'exclusions')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites2')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites3')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites4')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sitesn')
    @SectionLineParser.section_parser('moleculetype', 'position_restraints')
    @SectionLineParser.section_parser('moleculetype', 'pairs_nb')
    @SectionLineParser.section_parser('moleculetype', 'SETTLE')
    @SectionLineParser.section_parser('moleculetype', 'distance_restraints')
    @SectionLineParser.section_parser('moleculetype', 'orientation_restraints')
    @SectionLineParser.section_parser('moleculetype', 'dihedral_restraints')
    @SectionLineParser.section_parser('moleculetype', 'angle_restraints')
    @SectionLineParser.section_parser('moleculetype', 'angle_restraints_z')
    def _molecule(self, line, lineno=0):
        """
        Parses the lines of the [atoms] directive.
        """
        self.current_itp.append(line)

    def parse_define(self, line):
        """
        Parse define statemetns
        """
        tokens = line.split()

        if len(tokens) > 2:
            tag = line.split()[1]
            parameters = line.split()[2:]
        else:
            _, tag = line.split()
            parameters = True

        definition = {tag: parameters}
        self.topology.defines.update(definition)

    def parse_include(self, line):
        """
        parse include statemnts
        """
        path = line.split()[1].strip('\"')
        if self.current_meta:
           # the current file is between ifdef
           # however tag is not in defines we have
           # read so the file is not read
           if self.current_meta["condition"] == "ifdef"\
              and self.current_meta["tag"] not in self.topology.defines:
                 return
           # the current file is between ifndef
           # so if tag is defined we ignore this file
           elif self.current_meta["condition"] == "ifndef"\
              and self.current_meta["tag"] in self.topology.defines:
                 return

        if self.cwdir:
           filename = os.path.join(self.cwdir, path)
           cwdir = os.path.dirname(filename)
        else:
           cwdir = os.path.dirname(path)
           filename = path

        with open(filename, 'r') as _file:
            lines = _file.readlines()

        read_topology(lines, topology=self.topology, cwdir=cwdir)

    def _split_atoms_and_parameters(self, tokens, atom_idxs):
        """
        Returns atoms from line based on the indices defined in `atom_idxs`.
        It also interprets slices etc. stored as strings.

        Parameters:
        ------------
        tokens: collections.deque[str]
            Deque of token to inspect. The deque **can be modified** in place.
        atom_idxs: list of ints or strings that are valid python slices

        Returns:
        -----------
        list
        """

        atoms = []
        remove = []
        # first we extract the atoms from the indices given using
        # ints or slices
        for idx in atom_idxs:
            if isinstance(idx, int):
                atoms.append(tokens[idx])
                remove.append(idx)
            elif isinstance(idx, slice):
                atoms.extend(tokens[idx])
                idx_range = range(0, len(tokens))
                remove += idx_range[idx]
            else:
                raise IOError

        # everything that is left are parameters, which we
        # get by simply deleting the atoms from tokens

        for index in sorted(remove, reverse=True):
            del tokens[index]

        return atoms, tokens


def read_topology(lines, topology, cwdir=None):
    """
    Parses `lines` of itp format and adds the
    molecule as a block to `force_field`.

    Parameters
    ----------
    lines: list
        list of lines of an itp file
    force_field: :class:`vermouth.forcefield.ForceField`
    """
    director = TOPDirector(topology, cwdir)
    return list(director.parse(iter(lines)))



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Fixtures for tests
"""
import textwrap
import pytest
import numpy as np
import networkx as nx
import vermouth.forcefield
import vermouth.molecule
from vermouth.molecule import Block, Molecule
from polyply.src.meta_molecule import MetaMolecule

@pytest.fixture
def example_meta_molecule():
    """
    Example molecule with three residues each with
    different attributes for testing the assignment
    of links.

    Names:
    -------
    BB - BB1 - BB - BB1 - BB2 - BB - BB1
          |          |                |
         SC1        SC1              SC1
          |          |                |
         SC2        SC2              SC2

    Nodes:
    ------
     0  - 1    4  - 5 - 6   9 - 10
          |         |           |
          2         7           11
          |         |           |
          3         8           12
    """
    force_field = vermouth.forcefield.ForceField("test")
    block_A = Block(force_field=force_field)
    block_A.add_nodes_from([(0 , {'resid': 1,  'atomname': 'BB',  'atype': 'A', 'charge': 0.0, 'other': 'A', 'resname': 'A'}),
                            (1 , {'resid': 1,  'atomname': 'BB1', 'atype': 'B', 'charge': 0.0, 'other': 'A', 'resname': 'A'}),
                            (2 , {'resid': 1,  'atomname': 'SC1', 'atype': 'C', 'charge': 0.0, 'other': 'A', 'resname': 'A'}),
                            (3 , {'resid': 1,  'atomname': 'SC2', 'atype': 'D', 'charge': 1.0, 'other': 'A', 'resname': 'A'})])
    block_A.add_edges_from([(0, 1), (1, 2), (2, 3)])

    block_B = Block(force_field=force_field)
    block_B.add_nodes_from([(0 , {'resid': 1,  'atomname': 'BB',  'atype': 'A', 'charge': 0.0, 'resname': 'B'}),
                            (1 , {'resid': 1,  'atomname': 'BB1', 'atype': 'B', 'charge': 0.0, 'resname': 'B'}),
                            (2 , {'resid': 1,  'atomname': 'BB2', 'atype': 'A', 'charge': 0.0, 'resname': 'B'}),
                            (3 , {'resid': 1,  'atomname': 'SC1', 'atype': 'A', 'charge': -0.5, 'resname': 'B'}),
                            (4 , {'resid': 1,  'atomname': 'SC2', 'atype': 'C', 'charge': 0.5, 'resname': 'B'})])
    block_B.add_edges_from([(0, 1), (1, 2), (2, 3), (1, 4)])

    molecule = block_A.to_molecule()
    molecule.merge_molecule(block_B)
    molecule.merge_molecule(block_A)
    molecule.add_edges_from([(1, 4), (6, 9)])

    graph = MetaMolecule._block_graph_to_res_graph(molecule)
    meta_mol = MetaMolecule(graph, force_field=force_field, mol_name="test")
    meta_mol.molecule = molecule
    return meta_mol



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test linear algebra aux functions.
"""

import textwrap
import pytest
import numpy as np
import math
import networkx as nx
import vermouth
from vermouth.molecule import Interaction
import polyply
from polyply import TEST_DATA
from polyply.src.linalg_functions import center_of_geometry
from polyply.src.generate_templates import (find_atoms,
                                            _expand_inital_coords,
                                            _relabel_interaction_atoms,
                                            compute_volume, map_from_CoG,
                                            extract_block, GenerateTemplates,
					    find_interaction_involving)

class TestGenTemps:

      @staticmethod
      def test_find_atoms():
          G = nx.Graph()
          G.add_edges_from([(1, 2), (2, 3), (3, 4)])
          attrs = {1: {"resname": "test", "id": 'A'},
                   2: {"resname": "test", "id": 'B'},
                   3: {"resname": "testB", "id": 'A'},
                   4: {"resname": "testB", "id": 'B'}}
          nx.set_node_attributes(G, attrs)
          nodes = find_atoms(G, "resname", "test")
          assert nodes == [1, 2]
          nodes = find_atoms(G, "id", "A")
          assert nodes == [1, 3]

      @staticmethod
      def test_expand_inital_coords():
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P3 1 ALA SC1 2
        3 P2 1 ALA SC2 3
        4 VS 1 ALA SC3 3

        [ bonds ]
        1 2 1 0.2 100
        2 3 1 0.6 700

        [ virtual_sitesn ]
        4 2 3 1
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        polyply.src.polyply_parser.read_polyply(lines, ff)
        block = ff.blocks['GLY']
        coords = _expand_inital_coords(block)
        assert len(coords) == 4
        for pos in coords.values():
            assert len(pos) == 3

      @staticmethod
      def test_compute_volume():
          lines = """
          [ moleculetype ]
          GLY 1

          [ atoms ]
          1 P1 1 ALA BB 1
          2 P1 1 ALA SC1 2
          3 P1 1 ALA SC2 3
          4 P1 1 ALA SC3 3

          [ bonds ]
          1 2 1 0.2 100
          2 3 1 0.6 700
          3 4 1 0.2 700
          """
          meta_mol = polyply.MetaMolecule()
          nonbond_params = {frozenset(["P1", "P1"]): {"nb1": 0.47, "nb2":0.5}}

          lines = textwrap.dedent(lines).splitlines()
          ff = vermouth.forcefield.ForceField(name='test_ff')
          polyply.src.polyply_parser.read_polyply(lines, ff)
          block = ff.blocks['GLY']
          coords = _expand_inital_coords(block)
          vol = compute_volume(meta_mol, block, coords, nonbond_params)
          assert vol > 0.

      @staticmethod
      def test_map_from_CoG():
         lines = """
         [ moleculetype ]
         GLY 1
         [ atoms ]
         1 P4 1 ALA BB 1
         2 P3 1 ALA SC1 2
         3 P2 1 ALA SC2 3
         4 P2 1 ALA SC3 3
         [ bonds ]
         1 2 1 0.2 100
         2 3 1 0.6 700
         3 4 1 0.2 700
         """
         lines = textwrap.dedent(lines).splitlines()
         ff = vermouth.forcefield.ForceField(name='test_ff')
         polyply.src.polyply_parser.read_polyply(lines, ff)
         block = ff.blocks['GLY']
         coords = _expand_inital_coords(block)
         points = np.array(list(coords.values()))
         CoG = center_of_geometry(points)
         new_coords = map_from_CoG(coords)
         for coord in coords:
             coords[coord] == new_coords[coord] + CoG

      @staticmethod
      @pytest.mark.parametrize('atom_defs', (
      [1, 2, 3],
      (1, 2, 3)))
      def test_relabel_interaction(atom_defs):
         interaction = Interaction(atom_defs, ["1", "ga_2"], {"test": "value"})
         mapping = {1: "A", 2: "B", 3: "C"}
         new_interaction = _relabel_interaction_atoms(interaction, mapping)
         assert new_interaction.atoms == ["A", "B", "C"]
         assert new_interaction.parameters == ["1", "ga_2"]
         assert new_interaction.meta == {"test": "value"}

      @staticmethod
      def test_extract_block():
         lines = """
         [ moleculetype ]
         test 1
         [ atoms ]
         1 P4 1 GLY BB 1
         2 P3 1 GLY SC1 2
         3 P2 1 ALA SC2 3
         4 P2 1 ALA SC3 3
         [ bonds ]
         1 2 1 0.2 100
         2 3 1 0.6 700
         3 4 1 0.2 700
         [ moleculetype ]
         GLY 1
         [ atoms ]
         1 P4 1 GLY BB 1
         2 P3 1 GLY SC1 2
         [ bonds ]
         1 2 1 0.2 100
         """
         lines = textwrap.dedent(lines).splitlines()
         ff = vermouth.forcefield.ForceField(name='test_ff')
         polyply.src.polyply_parser.read_polyply(lines, ff)
         block = ff.blocks['test']
         molecule = block.to_molecule()
         new_block = extract_block(molecule, "GLY", {})
         for node in ff.blocks["GLY"]:
             atomname = ff.blocks["GLY"].nodes[node]["atomname"]
             assert ff.blocks["GLY"].nodes[node] == new_block.nodes[atomname]
         for inter_type in ff.blocks["GLY"].interactions:
             len(ff.blocks["GLY"].interactions[inter_type]) == len(new_block.interactions[inter_type])

      @staticmethod
      def test_run_molecule():
          top = polyply.src.topology.Topology.from_gmx_topfile(TEST_DATA + "/topology_test/system.top", "test")
          top.gen_pairs()
          top.convert_nonbond_to_sig_eps()
          GenerateTemplates(topology=top, max_opt=10).run_molecule(top.molecules[0])
          assert "PMMA" in top.volumes
          assert "PMMA" in top.molecules[0].templates

      @staticmethod
      @pytest.mark.parametrize('lines, result', (
        ("""
         [ moleculetype ]
         test 1
         [ atoms ]
         1 P4 1 GLY BB 1
         2 P3 1 GLY SC1 2
         3 P2 1 ALA SC2 3
         [ bonds ]
         1 2 1 0.2 100
         2 3 1 0.2 100
         """,
         (False, Interaction(atoms=["1", "2"], parameters=["1", "0.2", "100"], meta={}),
          "bonds")),
        ("""
         [ moleculetype ]
         test 1
         [ atoms ]
         1 P4 1 GLY BB 1
         2 P3 1 GLY SC1 2
         3 VS 1 ALA SC2 3
         4 P3 1 ALA SC3 3
         [ virtual_sitesn ]
         3 1 4 2 1
         """,
         (True, Interaction(atoms=["3", "4", "2", "1"], parameters=["1"], meta={}),
          "virtual_sitesn"))))
      def test_find_interaction_involving(lines, result):
          lines = textwrap.dedent(lines).splitlines()
          ff = vermouth.forcefield.ForceField(name='test_ff')
          polyply.src.polyply_parser.read_polyply(lines, ff)
          block = ff.blocks['test']
          result = find_interaction_involving(block, 1, 2)
          assert result == result

      @staticmethod
      @pytest.mark.parametrize('lines', (
        """
         [ moleculetype ]
         test 1
         [ atoms ]
         1 P4 1 GLY BB 1
         2 P3 1 GLY SC1 2
         3 P2 1 ALA SC2 3
         [ bonds ]
         1 2 1 0.2 100
         """,
        """
         [ moleculetype ]
         test 1
         [ atoms ]
         1 P4 1 GLY BB 1
         2 P3 1 GLY SC1 2
         3 VS 1 ALA SC2 3
         4 P3 1 ALA SC3 3
         """))
      def test_find_interaction_involving_error(lines):
          lines = textwrap.dedent(lines).splitlines()
          ff = vermouth.forcefield.ForceField(name='test_ff')
          polyply.src.polyply_parser.read_polyply(lines, ff)
          block = ff.blocks['test']
          with pytest.raises(IOError):
               find_interaction_involving(block, 1, 2)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import norm
"""
Rules for constructing virtual sites according
to GROMACS definitions.
"""

def vsn1(interaction, positions):
    """
    Construct a virtual-siten of type
    1 from interaction and positions.

    Parameters
    ---------
    interaction
    positions

    Returns
    -------
    np.ndarray(3)
    """
    weights = np.full((len(interaction.atoms[1:])), 1.)
    coords = np.array([positions[atom] for atom in interaction.atoms[1:]])
    vs_coord = np.average(coords, axis=0, weights=weights)
    return vs_coord

    #return coord

def vs2(interaction, positions):
    """
    Constructs virtual-site 2.

    Parameters
    ---------
    interaction
    positions

    Returns
    -------
    np.ndarray(3)
    """
    a = float(interaction.parameters[1])
    wi = 1 - a
    wj = a
    weights = [wi, wj]
    coords = np.array([positions[atom] for atom in interaction.atoms[1:]])
    vs_coord = np.average(coords, axis=0, weights=weights)
    return vs_coord

def vs3(interaction, positions):
    """
    Constructs virtual-site 3.

    Parameters
    ---------
    interaction
    positions

    Returns
    -------
    np.ndarray(3)
    """
    a = float(interaction.parameters[1])
    b = float(interaction.parameters[2])
    wi = 1 - a - b
    wj = a
    wk = b
    weights = [wi, wj, wk]
    coords = np.array([positions[atom] for atom in interaction.atoms[1:]])
    vs_coord = np.average(coords, axis=0, weights=weights)
    return vs_coord

def vs3fd(interaction, positions):
    """
    Construct virtual-site 3fd.

    Parameters
    ---------
    interaction
    positions

    Returns
    -------
    np.ndarray(3)
    """
    atoms = interaction.atoms[1:]
    r_i, r_j, r_k = positions[atoms[0]], positions[atoms[1]], positions[atoms[2]]
    r_ij = r_j - r_i
    r_jk = r_k - r_j
    r_ij = r_j - r_i
    a = float(interaction.parameters[1])
    b = float(interaction.parameters[2])

    return r_i + (b * (r_ij + a * r_jk)/norm(r_ij + a * r_jk))

def vs3fad(interaction, positions):
    """
    Construct virtual-site 3fad.

    Parameters
    ---------
    interaction
    positions

    Returns
    -------
    np.ndarray(3)
    """
    atoms = interaction.atoms[1:]
    r_i, r_j, r_k = positions[atoms[0]], positions[atoms[1]], positions[atoms[2]]
    theta, d = float(interaction.parameters[1]), float(interaction.parameters[2])
    r_ij = r_j - r_i
    r_jk = r_k - r_j
    r_ij = r_j - r_i
    r_normal = r_jk - r_ij * np.dot(r_ij, r_jk)/np.dot(r_ij, r_ij)
    angle1 = np.cos(np.deg2rad(theta)) * r_ij/norm(r_ij)
    angle2 = np.sin(np.deg2rad(theta)) * r_normal/norm(r_normal)
    vs = r_i + d * angle1 + d * angle2
    return vs

def vs3out(interaction, positions):
    """
    Construct virtual-site 3out.

    Parameters
    ---------
    interaction
    positions

    Returns
    -------
    np.ndarray(3)
    """
    atoms = interaction.atoms[1:]
    r_i, r_j, r_k = positions[atoms[0]], positions[atoms[1]], positions[atoms[2]]
    a = float(interaction.parameters[1])
    b = float(interaction.parameters[2])
    c = float(interaction.parameters[3])
    r_ij = r_j - r_i
    r_ik = r_k - r_i
    r_ij = r_j - r_i
    vs = r_i + float(a) * r_ij + float(b) * r_ik + float(c) * np.cross(r_ij, r_ik)
    return vs

def vs4fdn(interaction, positions):
    """
    Construct virtual-site 4fdn.

    Parameters
    ---------
    interaction
    positions

    Returns
    -------
    np.ndarray(3)
    """
    atoms = interaction.atoms[1:]
    a = float(interaction.parameters[1])
    b = float(interaction.parameters[2])
    c = float(interaction.parameters[3])
    r_i, r_j = positions[atoms[0]], positions[atoms[1]]
    r_k, r_l = positions[atoms[2]], positions[atoms[3]]
    r_ij = r_j - r_i
    r_ik = r_k - r_i
    r_il = r_l - r_i
    r_ja = a * r_ik - r_ij
    r_jb = b * r_il - r_ij
    rm = np.cross(r_ja, r_jb)
    vs = r_i + c * rm/norm(rm)
    return vs

def vserr(interaction, positions):
    raise IOError("Virtual-site 4 with function-type"
                  "1 is deprecated and cannot be used.")

# we don't differentiate between COM, COG
# for virtual_sitesn because the error
# is accaptable in terms of approximate
# structure generation

VIRTUAL_SITES = {('virtual_sitesn', '1'): vsn1,
                 ('virtual_sitesn', '2'): vsn1,
                 ('virtual_sitesn', '3'): vsn1,
                 ('virtual_sites4', '2'): vs4fdn,
                 ('virtual_sites4', '1'): vserr,
                 ('virtual_sites3', '4'): vs3out,
                 ('virtual_sites3', '3'): vs3fad,
                 ('virtual_sites3', '2'): vs3fd,
                 ('virtual_sites3', '1'): vs3,
                 ('virtual_sites2', '1'): vs2
                 }
def construct_vs(vs_type, interaction, positions):
    """
    Construct a vs from atoms in `positions` according
    to `vs_type` and atoms in `interaction`.
    """
    func = interaction.parameters[0]
    return VIRTUAL_SITES[(vs_type, func)](interaction, positions)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import vermouth
import polyply.src
from tqdm import tqdm

class Processor:
    """
    An abstract base class for processors. Subclasses must implement a
    `run_molecule` method.
    """
    def run_system(self, system):
        """
        Process `system`.
        Parameters
        ----------
        system: vermouth.system.System
            The system to process. Is modified in-place.
        """
        mols = []
        for molecule in tqdm(system.molecules):
            mols.append(self.run_molecule(molecule))
        system.molecules = mols

    def run_molecule(self, meta_molecule):
        """
        Process a single molecule. Must be implemented by subclasses.
        Parameters
        ----------
        molecule: vermouth.molecule.Molecule
            The molecule to process.
        Returns
        -------
        vermouth.molecule.Molecule
            Either the provided molecule, or a brand new one.
        """
        raise NotImplementedError




# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""
import math
import pytest
import numpy as np
from numpy.linalg import norm
import networkx as nx
import polyply
from polyply import TEST_DATA
from polyply.src.topology import Topology
from polyply.src.nonbond_engine import NonBondEngine
from polyply.src.random_walk import (fulfill_geometrical_constraints,
                                     pbc_complete,
                                     not_exceeds_max_dimensions,
                                     _take_step,
                                     RandomWalk,
                                     is_restricted)

@pytest.mark.parametrize('restraint_dict, point, result', (
    # test single geometrical constraint
    ({"restraints": [["in", np.array([0.0, 0.0, 0.0]), 1.0, "sphere"]]},
     np.array([0, 0, 0.5]),
     True
     ),
    ({"restraints": [["out", np.array([0.0, 0.0, 0.0]), 1.0, "sphere"]]},
     np.array([0, 0, 1.5]),
     True
     ),
    ({"restraints": [["in", np.array([0.0, 0.0, 0.0]), 1.0, "sphere"]]},
     np.array([0, 0, 1.5]),
     False
     ),
    ({"restraints": [["out", np.array([0.0, 0.0, 0.0]), 1.0, "sphere"]]},
     np.array([0.0, 0.0, 0.5]),
     False
     ),
    ({"restraints": [["in", np.array([0.0, 0.0, 0.0]), 1.0, 2.0, "cylinder"]]},
     np.array([0, 0.5, 0.5]),
     True
     ),
    ({"restraints": [["out", np.array([0.0, 0.0, 0.0]), 1.0, 2.0, "cylinder"]]},
     np.array([0, 1.5, 1.5]),
     True
     ),
    ({"restraints": [["in", np.array([0.0, 0.0, 0.0]), 1.0, 2.0, "cylinder"]]},
     np.array([0, 1.5, 1.5]),
     False
     ),
    ({"restraints": [["out", np.array([0.0, 0.0, 0.0]), 2.0, 2.0, 4.0, "rectangle"]]},
     np.array([0, 0.5, 0.5]),
     False
     ),
    ({"restraints": [["in", np.array([0.0, 0.0, 0.0]), 2.0, 2.0, 4.0, "rectangle"]]},
     np.array([0, 1.0, 0.5]),
     True
     ),
    ({"restraints": [["out", np.array([0.0, 0.0, 0.0]), 2.0, 2.0, 4.0, "rectangle"]]},
     np.array([0, 1.5, 4.5]),
     True
     ),
    ({"restraints": [["in", np.array([0.0, 0.0, 0.0]), 2.0, 2.0, 4.0, "rectangle"]]},
     np.array([0, 1.5, 4.5]),
     False
     ),
    ({"restraints": [["out", np.array([0.0, 0.0, 0.0]), 2.0, 2.0, 4.0, "rectangle"]]},
     np.array([0, 1.5, 3.9]),
     False
     ),
    # test default empty dict
    ({},
     np.array([0, 1.5, 3.9]),
     True
     ),
))
def test_geometric_restrictions(restraint_dict, point, result):
    assert fulfill_geometrical_constraints(point, restraint_dict) == result


@pytest.mark.parametrize('box_vect, point, result', (
    (np.array([5., 5., 10.]),
     np.array([0, 0, 0.5]),
     np.array([0, 0, 0.5])
     ),
    (np.array([5., 5., 10.]),
     np.array([5., 6., 0.5]),
     np.array([0., 1.0, 0.5])
    ),
    (np.array([5., 5., 10.]),
     np.array([5., -3., 0.5]),
     np.array([0., 2.0, 0.5])
     )))
def test_pbc_complete(box_vect, point, result):
    assert all(pbc_complete(point, box_vect) == result)


@pytest.mark.parametrize('box_vect, point, result', (
    (np.array([5., 5., 10.]),
     np.array([1., 1., 0.5]),
     True
     ),
    (np.array([5., 5., 10.]),
     np.array([5., 6., 0.5]),
     False
     )))
def test_not_exceeds_max_dimensions(box_vect, point, result):
    assert not_exceeds_max_dimensions(point, box_vect) == result


def test__take_step():
    coord = np.array([1.0, 1.0, 1.0])
    step_length = 0.5
    vectors = polyply.src.linalg_functions.norm_sphere(50)
    new_coord, _ = _take_step(
        vectors, step_length, coord, np.array([5.0, 5.0, 5.0]))
    assert math.isclose(norm(new_coord - coord), step_length)


@pytest.fixture
def nonbond_matrix():
    toppath = TEST_DATA + "/struc_build/system.top"
    topology = Topology.from_gmx_topfile(name="test", path=toppath)
    topology.preprocess()
    topology.volumes = {"PEO":0.43}
    return NonBondEngine.from_topology(topology.molecules,
                                       topology,
                                       box=np.array([10., 10., 10.]))
@pytest.fixture
def molecule():
    toppath = TEST_DATA + "/struc_build/system.top"
    topology = Topology.from_gmx_topfile(name="test", path=toppath)
    return topology.molecules[0]

def add_positions(nb_matrix, ncoords, pos=None):
    if isinstance(pos, type(None)):
        pos = np.array([[1.0, 1.0, 0.37],
                        [1.0, 1.0, 0.74],
                        [1.0, 1.0, 1.11],
                        [1.0, 1.0, 1.48],
                        [1.0, 1.0, 1.85],
                        [1.0, 1.0, 2.22],
                        [1.0, 1.0, 2.59],
                        [1.0, 1.0, 2.96],
                        [1.0, 1.0, 3.33],
                        [1.0, 1.0, 3.70],
                        [1.0, 1.0, 4.07]])

    nb_matrix.add_positions(pos[0], mol_idx=0, node_key=0, start=True)
    for idx, point in enumerate(pos[1:ncoords]):
        if all(point != np.array([np.inf, np.inf, np.inf])):
           nb_matrix.add_positions(point, mol_idx=0, node_key=idx+1, start=False)
    return nb_matrix

def test_rewind(nonbond_matrix):
    nb_matrix = add_positions(nonbond_matrix, 6)
    processor = RandomWalk(mol_idx=0, nonbond_matrix=nb_matrix, nrewind=3)
    # node 4 is already placed and hence is skipped over
    processor.placed_nodes = [(0, 0), (1, 1), (2, 2), (3, 3), (4, 5), (5, 6)]
    last_idx = processor._rewind(current_step=5)
    assert last_idx == 3
    for idx in [6, 5, 3]:
        assert all(nb_matrix.positions[idx] == np.array([np.inf, np.inf, np.inf]))

@pytest.mark.parametrize('new_point, result', (
    (np.array([1., 1., 2.96]),
     False
     ),
    (np.array([1., 1., 2.3]),
     True
     )))
def test_is_overlap(nonbond_matrix, molecule, new_point, result):
    nb_matrix = add_positions(nonbond_matrix, 6)
    proccessor = RandomWalk(mol_idx=0, nonbond_matrix=nb_matrix)
    proccessor.molecule = molecule
    # node 4 is already placed and hence is skipped over
    assert proccessor._is_overlap(new_point, 7, nrexcl=1) == result

@pytest.mark.parametrize('pos, expected', (
    # simple test; should just work
    (np.array([[1.0, 1.0, 0.37],
               [1.0, 1.0, 0.74],
               [1.0, 1.0, 1.11],
               [1.0, 1.0, 1.48],
               [1.0, 1.0, 1.85],
               [1.0, 1.0, 2.22],
               [1.0, 1.0, 2.59]]),
     True),
    # this will fail because all space is blocked
    (np.array([[1.0, 1.0, 0.67],
               [1.0, 1.0, 1.37],
               [1.0, 1.37, 1.0],
               [1.37, 1.0, 1.0],
               [1.0, 0.63, 1.0],
               [0.63, 1.0, 1.0],
               [1.0, 1.0, 1.0]]),
     False
     )))
def test_update_positions(nonbond_matrix, molecule, pos, expected):
    # add positions of the rest of the chain
    nb_matrix = add_positions(nonbond_matrix, 7, pos=pos)
    # create instance of processor
    proccessor = RandomWalk(mol_idx=0,
                            nonbond_matrix=nb_matrix,
                            maxdim=np.array([10., 10., 10.]),
                            max_force=100.0,
                            maxiter=49)
    # set molecule attribute which is normally set by the run_molecule class
    proccessor.molecule = molecule
    vector_bundle = polyply.src.linalg_functions.norm_sphere(50)
    status = proccessor.update_positions(vector_bundle=vector_bundle, current_node=7, prev_node=6)
    assert status == expected
    if status:
       assert all(nb_matrix.positions[7] != np.array([np.inf, np.inf, np.inf]))
    else:
       assert all(nb_matrix.positions[7] == np.array([np.inf, np.inf, np.inf]))

@pytest.mark.parametrize('build_attr, pos, start, npos', (
    # simple test; create all coordinates
    ({0: True, 1: True, 2: True, 3: True,
      4: True, 5: True, 6: True, 7: True, 8: True, 9: True},
     None,
     None,
     0),
    # start in the middle of the chain
    ({0: True, 1: True, 2: True, 3: True,
      4: True, 5: True, 6: True, 7: True, 8: True, 9: True},
     None,
     5,
     0),
    # here we look for a starting point and build the rest
    ({0: False, 1: False, 2: True, 3: True,
      4: True, 5: True, 6: True, 7: True, 8: True, 9: True},
     np.array([[1.0, 1.0, 0.67],
               [1.0, 1.0, 1.37]]),
     None,
     2,
     ),
    # here we need to skip one already defined position
    ({0: False, 1: True, 2: False, 3: True,
      4: True, 5: True, 6: True, 7: True, 8: True, 9: True},
     np.array([[1.0, 1.0, 0.67],
               [np.inf, np.inf, np.inf],
               [1.0, 1.0, 1.30]]),
     None,
     3),
    # here we trigger a rewind
   # ({0: False, 1: False, 2: False, 3: False,
   #   4: False, 5: False, 6: False, 7: True, 8: True, 9: True},
   #  np.array([[1.0, 1.0, 0.67],
   #            [1.0, 1.0, 1.37],
   #            [1.0, 1.37, 1.0],
   #            [1.37, 1.0, 1.0],
   #            [1.0, 0.63, 1.0],
   #            [0.63, 1.0, 1.0],
   #            [1.0, 1.0, 1.0],
   #            ]),
   #  None,
   #  7),
     ))
def test_run_molecule(nonbond_matrix, molecule, build_attr, npos, pos, start):
    # add positions of the rest of the chain
    nb_matrix = add_positions(nonbond_matrix, npos, pos=pos)
    # create instance of processor
    vector_bundle = polyply.src.linalg_functions.norm_sphere(500)
    proccessor = RandomWalk(mol_idx=0,
                            nonbond_matrix=nb_matrix,
                            maxdim=np.array([10., 10., 10.]),
                            max_force=100.0,
                            maxiter=49,
                            vector_sphere=vector_bundle,
                            start_node=start)
    # set molecule attribute which is normally set by the run_molecule class
    nx.set_node_attributes(molecule, build_attr, "build")
    proccessor.run_molecule(molecule)
    for pos in proccessor.nonbond_matrix.positions:
        assert all(pos != np.array([np.inf, np.inf, np.inf]))


@pytest.mark.parametrize('point, old_point, node_dict, expected', (
    # basic example
    (np.array([1., 1., 2.]),
     np.array([1., 1., 1.]),
     {"rw_options": [[np.array([0., 0., 1.0]), 90.0]]},
     True),
    # false because goes back
    (np.array([1., 1., 0.5]),
     np.array([1., 1., 1.]),
     {"rw_options": [[np.array([0., 0., 1.0]), 90.0]]},
     False),
    # false because angle not large enough
    (np.array([1.5, 1.5, 1.5]),
     np.array([1., 1., 1.]),
     {"rw_options": [[np.array([0., 0., 1.0]), 50.0]]},
     False),
))
def test_vector_push(point, old_point, node_dict, expected):
    status = is_restricted(point, old_point, node_dict)
    assert status == expected



import textwrap
import pytest
import numpy as np
import vermouth.forcefield
import vermouth.molecule
import polyply.src.ff_parser_sub

@pytest.mark.parametrize(
    'interaction_lines, edges, edge_attr', (
        (  # the edge section creates edges
            """
            [ edges ]
            SC1 SC2
            BB SC3
            """,
            (('SC1', 'SC2'), ('BB', 'SC3')),
            ({}, {}, {}),
        ),
        (  # the edge section creates edges and
           # can set edge attributes
            """
            [ edges ]
            SC1 SC2
            BB SC3 {"attr": "edge"}
            """,
            (('SC1', 'SC2'), ('BB', 'SC3')),
            ({}, {"attr": "edge"}),
        ),
    )
)
def test_interaction_edges(interaction_lines, edges, edge_attr):
    """
    Edges are created where expected.
    """
    lines = """
    [ moleculetype ]
    GLY 1
    [ atoms ]
    1 P4 1 ALA BB 1
    2 P3 1 ALA SC1 2
    3 P2 1 ALA SC2 3
    4 P2 1 ALA SC3 3
    """
    lines = textwrap.dedent(lines) + textwrap.dedent(interaction_lines)
    lines = lines.splitlines()

    ff = vermouth.forcefield.ForceField(name='test_ff')
    polyply.src.ff_parser_sub.read_ff(lines, ff)
    block = ff.blocks['GLY']
    for edge, attr in zip(edges, edge_attr):
        assert block.has_edge(edge[0], edge[1]) or block.has_edge(edge[1], edge[0])
        assert attr == block.edges[edge]

def test_interaction_edges_error():
    """
    Make sure an error is raised when someone
    tries to set node attributes in edges.
    """
    lines = """
    [ moleculetype ]
    GLY 1
    [ atoms ]
    1 P4 1 ALA BB 1
    2 P3 1 ALA SC1 2
    3 P2 1 ALA SC2 3
    4 P2 1 ALA SC3 3
    [ edges ]
    BB SC1
    SC1 {"some_attr": "some_value"} SC2
    """
    lines = textwrap.dedent(lines)
    lines = lines.splitlines()

    ff = vermouth.forcefield.ForceField(name='test_ff')
    with pytest.raises(IOError):
        polyply.src.ff_parser_sub.read_ff(lines, ff)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import numpy as np
from numpy.linalg import norm
import networkx as nx
from .processor import Processor
from .linalg_functions import norm_sphere
from .linalg_functions import _vector_angle_degrees
from .graph_utils import neighborhood
"""
Processor implementing a random-walk to generate
coordinates for a meta-molecule.
"""


def pbc_complete(point, maxdim):
    """
    Wrap point around pbc conditions to keep
    points from being larger than the compontents of
    the maxdim vector. Note that all coodinates in
    polyply are definiet positive.

    Parameters:
    -----------
    point: np.ndarray
    maxdim: np.ndarray

    Returns:
    --------
    np.ndarray
    """
    return point % maxdim


def _take_step(vectors, step_length, coord, box):
    """
    Given a list of unit `vectors` choose one randomly,
    multiply it by the `step_length` and add to `coord`.

    Parameters
    ----------
    vectors: list[np.ndarray(n, 3)]
    step_length: float
    coord: np.ndarray(3)

    Returns
    -------
    np.ndarray(3)
        the new coordinate
    int
        the index of the vector choosen
    """
    index = random.randint(0, len(vectors) - 1)
    new_coord = coord + vectors[index] * step_length
    new_coord = pbc_complete(new_coord, box)
    return new_coord, index


def not_exceeds_max_dimensions(point, maxdim):
    """
    Check if point is within the compontents of
    the maxdim vector. Note that all coodinates in
    polyply are definiet positive.

    Parameters:
    -----------
    point: np.ndarray
    maxdim: np.ndarray

    Returns:
    --------
    bool
    """
    return np.all(point < maxdim) and np.all(point > np.array([0., 0., 0.]))

def is_restricted(point, old_point, node_dict):
    """
    The function checks, if the step `old_point` to
    `point` is in a direction as defined by a plane
    with normal and angle as set in the `node_dict`
    options with keyword rw_options. It is true
    if the direction vector point-old_point is
    pointing some max angle from the normal of the plane
    in the same direction as an angle specifies.
    To check we compute the signed angle of the vector
    `point`-`old_point` has with the plane defined
    by a normal and compare to the reference angle.
    The angle needs to have the same sign as angle
    and not be smaller in magnitude.

    Parameters:
    -----------
    point: np.ndarray
    old_point: np.ndarray
    node_dict:  dict["rw_options"][[np.ndarray, float]]
        dict with key rw_options containing a list of lists
        specifying a normal and an angle

    Returns:
    -------
    bool
    """

    if not "rw_options" in node_dict:
        return True

    normal, ref_angle = node_dict["rw_options"][0]
    # check condition 1
    sign = np.sign(np.dot(normal, point - old_point))
    if sign != np.sign(ref_angle):
        return False

    # check condition 2
    angle = _vector_angle_degrees(normal, point - old_point)
    if angle > np.abs(ref_angle):
        return False
    return True


def in_cylinder(point, parameters):
    """
    Assert if a point is within a cylinder or outside a
    cylinder as defined in paramters. Note the cylinder
    is z-aligned always.

    Parameters:
    -----------
    point: np.ndarray
        reference point
    parameters: abc.iteratable

    Returns:
    --------
    bool
    """
    in_out = parameters[0]
    diff = parameters[1] - point
    radius = norm(diff[:2])
    half_heigth = diff[2]
    if in_out == "in" and radius < parameters[2] and np.abs(half_heigth) < parameters[3]:
        return True
    elif in_out == "out" and (radius > parameters[2] or half_heigth > np.abs(parameters[3])):
        return True
    else:
        return False

def in_rectangle(point, parameters):
    """
    Assert if a point is within a rectangle or outside a
    rectangle as defined in paramters:

    Parameters:
    -----------
    point: np.ndarray
        reference point
    parameters: abc.iteratable

    Returns:
    --------
    bool
    """
    in_out = parameters[0]
    diff = parameters[1] - point
    check = [np.abs(dim) < max_dim for dim,
             max_dim in zip(diff, parameters[2:])]
    if in_out == "in" and not all(check):
        return False
    elif in_out == "out" and all(check):
        return False
    else:
        return True


def in_sphere(point, parameters):
    """
    Assert if a point is within a sphere or outside a
    sphere as defined in paramters:

    Parameters:
    -----------
    point: np.ndarray
        reference point
    parameters: abc.iteratable

    Returns:
    --------
    bool
    """
    in_out = parameters[0]
    r = norm(parameters[1] - point)
    if in_out == 'in' and r > parameters[2]:
        return False
    elif in_out == 'out' and r < parameters[2]:
        return False
    else:
        return True


# methods of geometrical restraints known to polyply
RESTRAINT_METHODS = {"cylinder": in_cylinder,
                     "rectangle": in_rectangle,
                     "sphere":  in_sphere}


def fulfill_geometrical_constraints(point, node_dict):
    """
    Assert if a point fulfills a geometrical constraint
    as defined by the "restraint" key in a dictionary.
    If there is no key "restraint" the function returns true.

    Parameters:
    -----------
    point: np.ndarray
        reference point

    node_dict: :class:ditc

    Returns:
    --------
    bool
    """
    if "restraints" not in node_dict:
        return True

    for restraint in node_dict['restraints']:
        restr_type = restraint[-1]
        if not RESTRAINT_METHODS[restr_type](point, restraint):
            return False

    return True


def _find_starting_node(meta_molecule):
    """
    Find the first node that has coordinates if there is
    otherwise return first node in list of nodes.
    """
    for node in meta_molecule.nodes:
        if "build" not in meta_molecule.nodes[node]:
            return node
    return next(iter(meta_molecule.nodes()))


class RandomWalk(Processor):
    """
    Add coordinates at the meta_molecule level
    through a random walk for all nodes which have
    build defined as true.
    """
    def __init__(self,
                 mol_idx,
                 nonbond_matrix,
                 start=np.array([0, 0, 0]),
                 step_fudge=0.8,
                 maxiter=80,
                 maxdim=None,
                 max_force=1e3,
                 vector_sphere=norm_sphere(5000),
                 start_node=None,
                 nrewind=5):

        self.mol_idx = mol_idx
        self.nonbond_matrix = nonbond_matrix
        self.start = start
        self.maxiter = maxiter
        self.maxdim = maxdim
        self.vector_sphere = vector_sphere
        self.success = False
        self.max_force = max_force
        self.step_fudge = step_fudge
        self.start_node = start_node
        self.nrewind = nrewind
        self.placed_nodes = []

    def _rewind(self, current_step):
        nodes = [node for _, node in self.placed_nodes[-self.nrewind:-1]]
        self.nonbond_matrix.remove_positions(self.mol_idx, nodes)
        step_count = self.placed_nodes[-self.nrewind][0]
        self.placed_nodes = self.placed_nodes[:-self.nrewind]
        return step_count

    def _is_overlap(self, point, node, nrexcl=1):
        neighbours = neighborhood(self.molecule, node, nrexcl)
        force_vect = self.nonbond_matrix.compute_force_point(point,
                                                             self.mol_idx,
                                                             node,
                                                             neighbours,
                                                             potential="LJ")
        return norm(force_vect) > self.max_force

    def update_positions(self, vector_bundle, current_node, prev_node):
        """
        Take an array of unit vectors `vector_bundle` and generate the coordinates
        for `current_node` by adding a random vector to the position of the previous
        node `prev_node`. The length of that vector is defined as 2 times the vdw-radius
        of the two nodes. The position is updated in place.

        Parameters
        ----------
        vector_bunde: np.ndarray(m,3)
        meta_molecule: :class:`polyply.src.meta_molecule.MetaMolecule`
        current_node: node_key[int, str]
        prev_node: node_key[int, str]
        topology: :class:`polyply.src.topology.Topology`
        maxiter: int
           maximum number of iterations
        """
        last_point = self.nonbond_matrix.get_point(self.mol_idx, prev_node)
        step_length = self.step_fudge * self.nonbond_matrix.get_interaction(self.mol_idx,
                                                                            self.mol_idx,
                                                                            prev_node,
                                                                            current_node)[0]
        step_count = 0
        while True:

            new_point, index = _take_step(vector_bundle, step_length, last_point, self.maxdim)

            if fulfill_geometrical_constraints(new_point, self.molecule.nodes[current_node])\
                and is_restricted(new_point, last_point, self.molecule.nodes[current_node])\
                and not self._is_overlap(new_point, current_node):

                    self.nonbond_matrix.add_positions(new_point,
                                                      self.mol_idx,
                                                      current_node,
                                                      start=False)
                    return True

            elif step_count == self.maxiter:
                return False

            else:
                step_count += 1
                vector_bundle = np.delete(vector_bundle, index, axis=0)

    def _random_walk(self, meta_molecule):
        """
        Perform a random_walk to build positions for a meta_molecule, if
        no position is present for an atom.

        Parameters
        ----------
        meta_molecule:  :class:`polyply.src.meta_molecule.MetaMolecule`
        """
        if not self.start_node:
            first_node = _find_starting_node(meta_molecule)
        else:
            first_node = self.start_node

        if "position" not in meta_molecule.nodes[first_node]:
            constrained = fulfill_geometrical_constraints(self.start,
                                                          self.molecule.nodes[first_node])

            if constrained and not self._is_overlap(self.start, first_node):
                self.nonbond_matrix.add_positions(self.start,
                                                  self.mol_idx,
                                                  first_node,
                                                  start=True)
                self.success = True
            else:
                self.success = False
                return

        vector_bundle = self.vector_sphere.copy()
        count = 0
        path = list(nx.bfs_edges(meta_molecule, source=first_node))
        step_count = 0

        while step_count < len(path):
            prev_node, current_node = path[step_count]

            if not meta_molecule.nodes[current_node]["build"]:
                step_count += 1
                continue

            status = self.update_positions(vector_bundle,
                                           current_node,
                                           prev_node)
            self.success = status
            self.placed_nodes.append((step_count, current_node))

            # in principle this count here is not needed, however,
            # we need to check the performance in terms of strucutre
            # generation before doing any adjustments here
            if not self.success and count < self.maxiter:
                if len(self.placed_nodes) < self.nrewind+1:
                    return
                step_count = self._rewind(step_count)
            elif not self.success:
                return
            else:
                count = 0
                step_count += 1

            count += 1

    def run_molecule(self, meta_molecule):
        """
        Perform the random walk for a single molecule.
        """
        self.molecule = meta_molecule
        self._random_walk(meta_molecule)
        return meta_molecule



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test backmapping
"""
import numpy as np
from numpy.linalg import norm
import networkx as nx
import vermouth
from vermouth.graph_utils import make_residue_graph
from polyply import MetaMolecule
from polyply.src.backmap import Backmap

def test_backmapping():
    """
    Integral test for the backmapping processor.
    """
    molecule = vermouth.molecule.Molecule()
    molecule.add_edges_from([(1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7)])
    nx.set_node_attributes(molecule, {
        1: {"resname": "test", "resid": 1, "atomname": "A"},
        2: {"resname": "test", "resid": 1, "atomname": "B"},
        3: {"resname": "test", "resid": 1, "atomname": "C"},
        4: {"resname": "test", "resid": 2, "atomname": "A"},
        5: {"resname": "test", "resid": 2, "atomname": "B"},
        6: {"resname": "test", "resid": 2, "atomname": "C"},
        7: {"resname": "test", "resid": 3, "atomname": "C",
            "position": np.array([4., 4., 4.])}
        })

    graph = make_residue_graph(molecule)
    meta_molecule = MetaMolecule(graph)
    meta_molecule.molecule = molecule

    nx.set_node_attributes(meta_molecule, {0: {"resname": "test",
                                               "position": np.array([0, 0, 0]),
                                               "resid": 1, "build": True},
                                           1: {"resname": "test",
                                               "position": np.array([0, 0, 1.0]),
                                               "resid": 2, "build": True},
                                           2: {"resname": "test",
                                               "position": np.array([0, 0, 2.0]),
                                               "resid": 3, "build": False}})
    # test if disordered template works
    meta_molecule.templates = {"test": {"B": np.array([0, 0, 0]),
                                        "A": np.array([0, 0, 0.5]),
                                        "C": np.array([0, 0.5, 0])}}

    Backmap().run_molecule(meta_molecule)
    for node in meta_molecule.molecule.nodes:
        assert "position" in meta_molecule.molecule.nodes[node]
    assert norm(meta_molecule.molecule.nodes[7]["position"] - np.array([4., 4., 4.])) == 0



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test constructions of virutal-sites. Reference values come from
a GROMACS energy minimization.
"""

import numpy as np
import pytest
from vermouth.molecule import Interaction
from ..src.virtual_site_builder import construct_vs


@pytest.mark.parametrize('vs_type, interaction, positions, result', (
    # vsn
    ("virtual_sitesn",
     Interaction(atoms=[6, 1, 2, 3, 4, 5], parameters=["1"], meta={}),
     {1: np.array([4.924, 0.353, 0.036]), 2: np.array([5.191, 0.023, 0.208]),
      3: np.array([5.190, -0.231, -0.162]), 4: np.array([4.734, -0.268, -0.131]),
      5: np.array([4.548, 0.088, 0.046])},
     np.array([4.917, -0.007, -0.001])
     ),
    #  vs2
    ("virtual_sites2",
     Interaction(atoms=[3, 1, 2], parameters=["1", "0.5"], meta={}),
     {1: np.array([0.000, 0.000, 0.000]), 2: np.array([-0.012, 0.002, 0.024])},
     np.array([-0.006, 0.001, 0.012])
     ),
    #  vs31
    ("virtual_sites3",
     Interaction(atoms=[4, 1, 2, 3], parameters=[
                 "1", "0.223", "0.865"], meta={}),
     {1: np.array([4.717, 0.333, 4.974]), 2: np.array([4.983, -0.028, 5.098]),
      3: np.array([5.307, -0.324, 4.951])},
     np.array([5.287, -0.316, 4.982])
     ),
    #  vs32
    ("virtual_sites3",
        Interaction(atoms=[4, 1, 2, 3], parameters=[
                    "2", "0.223", "0.865"], meta={}),
        {1: np.array([4.717, 0.333, 4.974]), 2: np.array([4.983, -0.028, 5.098]),
         3: np.array([5.307, -0.324, 4.951])},
        np.array([5.247, -0.336, 5.117])
     ),
    #  vs33
    ("virtual_sites3",
        Interaction(atoms=[4, 1, 2, 3], parameters=[
                    "3", "150", "0.22"], meta={}),
        {1: np.array([4.717, 0.333, 4.974]), 2: np.array([4.983, -0.028, 5.098]),
         3: np.array([5.307, -0.324, 4.951])},
        np.array([4.652, 0.479, 4.823])
     ),
    #  vs34
    ("virtual_sites3",
     Interaction(atoms=[4, 1, 2, 3], parameters=[
                 "4", "0.2", "0.3", "1.5"], meta={}),
     {1: np.array([4.717, 0.333, 4.974]), 2: np.array([4.983, -0.028, 5.098]),
      3: np.array([5.307, -0.324, 4.951])},
     np.array([5.081, 0.182, 5.049])
     ),
    # vs4
    ("virtual_sites4",
        Interaction(atoms=[5, 1, 2, 3, 4], parameters=[
                    "2", "0.22", "0.45", "0.6"], meta={}),
        {1: np.array([4.917, 0.223, 0.192]), 2: np.array([5.064, 0.143, -0.214]),
         3: np.array([5.185, -0.244, 0.044]), 4: np.array([4.730, -0.149, -0.012])},
        np.array([5.156, -0.327, 0.215])
     )
))
def test(vs_type, interaction, positions, result):
    """
    Test if the six virtual sites are constructed properly.
    """
    vs_coord = construct_vs(vs_type, interaction, positions)
    assert all(np.isclose(vs_coord, result, atol=10**-3.))


def test_IOError():
    """
    Tests if VS1 with function type 1 raises error.
    """
    with pytest.raises(IOError):
        construct_vs("virtual_sites4", Interaction(
            atoms=[], parameters=["1"], meta={}), {})



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import os.path
from setuptools import find_packages, setup
import codecs
# this comes form here https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

setup(
    package_data={'': package_files('polyply/data')
                  + package_files('polyply/tests/test_data'),},
    scripts=['bin/polyply'],
    pbr=True,
    version=get_version("polyply/__init__.py")
)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import vermouth
from vermouth.gmx.rtp import read_rtp
from vermouth.citation_parser import read_bib
from polyply import DATA_PATH
from .ff_parser_sub import read_ff
from .polyply_parser import read_polyply

FORCE_FIELD_PARSERS = {'rtp': read_rtp, 'ff': read_ff, 'itp': read_polyply, 'bib': read_bib}


def _resolve_lib_paths(lib_names, data_path):
    """
    select the appropiate files from data_path
    according to library names given.

    Parameters
    ----------
    lib_names: list[`os.path`]
        list of library names
    data_path:
        location of library files
    """
    files = []
    for name in lib_names:
        directory = os.path.join(data_path, name)
        for _file in os.listdir(directory):
            _file = os.path.join(directory, _file)
            files.append(_file)
    return files


def read_ff_from_files(paths, force_field):
    """
    read the input files for the defintion of blocks and links.

    Parameters
    ----------
    paths: list
           List of vaild file paths
    force_field: class:`vermouth.force_field.ForceField`

    Returns
    -------
    force_field: class:`vermouth.force_field.ForceField`
       updated forcefield

    """

    def wrapper(parser, path, force_field):
        with open(path, 'r') as file_:
            lines = file_.readlines()
            parser(lines, force_field=force_field)

    for path in paths:
        path = pathlib.Path(path)
        file_extension = path.suffix.casefold()[1:]

        try:
            parser = FORCE_FIELD_PARSERS[file_extension]
            wrapper(parser, path, force_field)
        except KeyError:
            raise IOError(
                "Cannot parse file with extension {}.".format(file_extension))

    return force_field


def load_library(name, lib_names, extra_files):
    """
    Load libraries and extra-files into vermouth
    force-field.

    Parameters
    ----------
    name: str
      Force-field name
    lib_names: list
      List of lirbary names
    extra_files: list
      List of extra files to include

    Returns
    -------
    :class:`vermouth.forcefield.Forcefield`
    """
    force_field = vermouth.forcefield.ForceField(name)
    if lib_names and extra_files:
        lib_files = _resolve_lib_paths(lib_names, DATA_PATH)
        all_files = lib_files + extra_files
    elif lib_names:
        all_files = _resolve_lib_paths(lib_names, DATA_PATH)
    elif extra_files:
        all_files = extra_files

    read_ff_from_files(all_files, force_field)
    return force_field



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
High level API for the polyply itp generator
"""
import sys
from pathlib import Path
import networkx as nx
import vermouth
import vermouth.forcefield
from vermouth.log_helpers import StyleAdapter, get_logger
from vermouth.file_writer import open, DeferredFileWriter
from vermouth.citation_parser import citation_formatter
import polyply
import polyply.src.polyply_parser
from polyply import (DATA_PATH, MetaMolecule, ApplyLinks, Monomer, MapToMolecule)
from .load_library import load_library

LOGGER = StyleAdapter(get_logger(__name__))

def split_seq_string(sequence):
    """
    Split a string definition for a linear sequence into monomer
    blocks and raise errors if the sequence is not valid.
    Parameters
    -----------
    sequence: str
            string of residues format name:number

    Returns:
    ----------
    list
       list of `polyply.Monomers`
    """
    raw_monomers = sequence
    monomers = []
    for monomer in raw_monomers:
        resname, n_blocks = monomer.split(":")
        n_blocks = int(n_blocks)
        monomers.append(Monomer(resname=resname, n_blocks=n_blocks))
    return monomers

def gen_params(args):
    # Import of Itp and FF files
    LOGGER.info("reading input and library files",  type="step")
    force_field = load_library(args.name, args.lib, args.inpath)

    # Generate the MetaMolecule
    if args.seq:
       LOGGER.info("reading sequence from command",  type="step")
       monomers = split_seq_string(args.seq)
       meta_molecule = MetaMolecule.from_monomer_seq_linear(monomers=monomers,
                                                            force_field=force_field,
                                                            mol_name=args.name)
    #ToDo
    # fix too broad except
    elif args.seq_file:
       LOGGER.info("reading sequence from file",  type="step")
       extension = args.seq_file.suffix.casefold()[1:]
       try:
           parser = getattr(MetaMolecule, 'from_{}'.format(extension))
           meta_molecule = parser(json_file=args.seq_file,
                                  force_field=force_field,
                                  mol_name=args.name)
       except AttributeError:
         raise IOError("Cannot parse file with extension {}.".format(extension))

    # Do transformationa and apply link
    LOGGER.info("mapping sequence to molecule",  type="step")
    meta_molecule = MapToMolecule(force_field).run_molecule(meta_molecule)
    LOGGER.info("applying links between residues",  type="step")
    meta_molecule = ApplyLinks().run_molecule(meta_molecule)

    # Raise warning if molecule is disconnected
    if not nx.is_connected(meta_molecule.molecule):
        n_components = len(list(nx.connected_components(meta_molecule.molecule)))
        msg = "You molecule consists of { } disjoint parts. Perhaps links were not applied correctly."
        LOGGER.warning(msg, n_components)

    with open(args.outpath, 'w') as outpath:
        header = [ ' '.join(sys.argv) + "\n" ]
        header.append("Please cite the following papers:")
        for citation in meta_molecule.molecule.citations:
            cite_string =  citation_formatter(meta_molecule.molecule.force_field.citations[citation])
            LOGGER.info("Please cite: " + cite_string)
            header.append(cite_string)

        vermouth.gmx.itp.write_molecule_itp(meta_molecule.molecule, outpath,
                                            moltype=args.name, header=header)
    DeferredFileWriter().write()

# ducktape for renaming the itp tool
gen_itp = gen_params



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test graph related functions
"""
import pytest
import networkx as nx
import polyply
from .example_fixtures import example_meta_molecule

@pytest.mark.parametrize('source, max_length, min_length, expected',(
                        (4, 1, 1, [4, 1, 9, 10]),
                        (4, 2, 1, [4, 1, 9, 10, 0, 3]),
                        (4, 3, 3, [0, 3, 7, 8, 2]),
                        (0, 1, 1, [0, 1, 2])
                        ))
def test_neighbourhood(source, max_length, min_length, expected):
    graph = nx.balanced_tree(r=2, h=3)
    neighbours = polyply.src.graph_utils.neighborhood(graph,
                                                      source,
                                                      max_length,
                                                      min_length=min_length)
    assert set(neighbours) == set(expected)

@pytest.mark.parametrize('edges, expected',(
                        # simple linear
                        ([(0, 1), (1, 2), (2, 3)], False),
                        # simple cyclic
                        ([(0, 1), (1, 2), (2, 3), (3, 0)], False),
                        # simple branched
                        ([(0, 1), (1, 2), (1, 3), (3, 4)], True),
                        # cyclic branched
                        ([(0, 1), (1, 2), (2, 3), (3, 0), (0, 5)], True),
                        # no nodes
                        ([], False)
                        ))
def test_is_branched(edges, expected):
    graph = nx.Graph()
    graph.add_edges_from(edges)
    result = polyply.src.graph_utils.is_branched(graph)
    assert result == expected

@pytest.mark.parametrize('nodes, expected',(
                        ((0, 1), [(1, 4)]),
                        # central residue
                        ((1, 2), [(6, 9)]),
                        ))
def test_find_connecting_edges(example_meta_molecule, nodes, expected):
    result =  polyply.src.graph_utils.find_connecting_edges(example_meta_molecule,
                                                            example_meta_molecule.molecule,
                                                            nodes)
    assert result == expected



# -*- coding: utf-8 -*-
# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
from pylint import lint


if __name__ == '__main__':
    desc = "PyLint wrapper that add the --fail-under option."\
           " All other arguments are passed to pylint."
    parser = argparse.ArgumentParser(description=desc, allow_abbrev=False)
    parser.add_argument('--fail-under', dest='threshold', type=float, default=8,
                        help='If the final score is more than THRESHOLD, exit with'
                        ' exitcode 0, and pylint\'s exitcode otherwise.')

    args, remaining_args = parser.parse_known_args()

    threshold = args.threshold
    run = lint.Run(remaining_args, do_exit=False)
    score = run.linter.stats['global_note'] # Yes this is a terrible name for the score
    if score < threshold:
        sys.exit(run.linter.msg_status)
